/* ----------------------------------------------------------------------------
 *  Serial Communication Lib							
 * ----------------------------------------------------------------------------
 *  prog: max.rheiner@zhdk.ch / Zhdk / Interaction Design
 *  date: 11/16/2011
 * ----------------------------------------------------------------------------
 */

import java.lang.reflect.Method;

public class SerialComLib
{
  final static int Ok = 0;
  final static int Error_Command = 1;
  final static int Error_Checksum = 2;
  final static int Error_Timeout = 3;
  final static int Error_DataHeader = 4;
  final static int Error_Data = 5;
  final static int Error_DataSize = 6;
  final static int Error_NoData = 7;

  final static int MAX_BUFFER_SIZE = 255;
  final static int MIN_PACKET_LENGTH = 5; //2(header) + 1(commandId) + 1(length) + 1(checksum) = 5
  final static int DATAHEADER1 = 0x1a;
  final static int DATAHEADER2 = 0xd0;

  protected int     _bufferLength;
  protected int     _bufferPos;
  protected int[]   _buffer;

  protected int    _cmdId;

  protected Serial   _serial;
  protected PApplet  _parent;

  protected Method  _onReceiveSerialComMethod;

  public SerialComLib(PApplet parent, String serialPort, int baudRate)
  {
    _parent = parent;
    _serial = new Serial(parent, serialPort, baudRate) ;

    _buffer = new int[MAX_BUFFER_SIZE];
    _bufferLength = 0;
    _bufferPos = 0;

    _cmdId = 0;

    _onReceiveSerialComMethod = getMethodRef(_parent,"onReceiveSerialCom", new Class[] { int.class,SerialComLib.class });
  }

  public void update()
  {
    if(readPacket() == Ok)
    {
      try 
      {
        _onReceiveSerialComMethod.invoke(_parent, new Object[] {_cmdId, this});
      } 
      catch (Exception e) 
      {
      }
    }  
  } 

  public Method getMethodRef(Object obj, String methodName, Class[] paraList)
  {
    Method	ret = null;
    try {
      ret = obj.getClass().getMethod(methodName, paraList);
    } 
    catch (Exception e) 
    { // no such method, or an error.. which is fine, just ignore
    }
    return ret;
  }

  // send packet

  public void beginSend(int cmdId)
  {
    _bufferPos = 0;
    _cmdId = cmdId;
  }

  public void sendByte(byte data)
  {
    _buffer[_bufferPos++] = data;
  }

  public void sendByte(int data)
  {
    _buffer[_bufferPos++] = data;
  }

  public void sendUShort(int data)
  {
    _buffer[_bufferPos++] = byte(0xff & data);
    _buffer[_bufferPos++] = byte((0xff00 & data) >> 8);
  }

  public void sendFloat(float data)
  {
  }

  public void endSend()
  {
    writeCmd(_cmdId, _bufferPos, _buffer);
  }

  protected byte checksum(int len, int[] data)
  {
    int checkSum = 0;
    for (int i=0;i < len;i++)
      checkSum += data[i];
    return((byte)(0xff - (0xff & checkSum)));
  }

  protected boolean testChecksum(int len, int[] data, int checksumVal)
  {
    return checksum(len, data) == checksumVal;
  }


  protected void writeCmd(int cmdId, int dataLen, int[] data)
  {
    // write header
    _serial.write((byte)DATAHEADER1);
    _serial.write((byte)DATAHEADER2);

    // write commandId
    _serial.write((byte)cmdId);

    // write data
    _serial.write((byte)dataLen);
    if (dataLen == 0)
      _serial.write((byte)0xff);  // checksum
    else
    {
      for (int i=0;i<dataLen;i++)
        _serial.write((byte)data[i]);

      // calc the checksume and write it
      _serial.write(checksum(dataLen, data));
    }
  } 

  
   // readn packet
   int readPacket()
   {
     if(available() == false)
       return Error_DataHeader;
    

    // debug
    /*
     int count = _serial.available();
     _parent.println("_serial.available(): " + _serial.available());
     for(int i=0;i < count;i++)
       _parent.print(hex(_serial.read()) + ",");
     println();
     return Error_DataHeader;
   */
      if(findDataHeader() == false)
        return Error_DataHeader;
        
      _cmdId = 0;
      _bufferLength = 0;
      _bufferPos = 0;
      return readCmd();          
   }
   
   int getCmd()
   {
    return _cmdId;
   }   
   
   int getCmdLength()
   {
    return _bufferLength;
   }
   
   int getByte()
   {
      if(_bufferPos < _bufferLength)
          return _buffer[_bufferPos++];
      return 0;   
   }
   
   int getUShort()
   {
     if(_bufferPos+1 < _bufferLength)
     {     
        return((_buffer[_bufferPos++] & 0xff) + ((_buffer[_bufferPos++] & 0xff) << 8 ));
     }
     return 0;  
   }
   
   float getFloat()
   {
     return 0;
   }
   
   boolean isPacketEnd()
   {
     return(_bufferPos >= _bufferLength);
   }
   
   int[] getData()
   {
     return _buffer;
   }
   
   
   public boolean available()
   {
     return (_serial.available() >= MIN_PACKET_LENGTH)? true : false;
   }
   
   
   public int readCmd()
   {
     _bufferPos = 0;
     _bufferLength =0;
     
     // read cmd id
     _cmdId = _serial.read();
     
     // read data length
     _bufferLength = _serial.read();  
     
     // read data
     if(_bufferLength < _buffer.length)
     {
       for(int i=0;i<_bufferLength;i++)
         _buffer[i] = (byte)_serial.read();
     }       
     else
       return Error_DataSize;
     /*
     _parent.println("_cmdId: " + hex(_cmdId));
     _parent.println("_bufferLength: " + _bufferLength);
     */
     
     // read checksum
     if(testChecksum(_bufferLength,_buffer,(byte)_serial.read()))
       return Ok;

     return Error_Checksum;
   }
   
   protected boolean findDataHeader()
   { 
     if(_serial.read() == DATAHEADER1)
     {
       if(_serial.read() == DATAHEADER2)
         return true;
     }
    return false;
  }
  
 
}

